import Quill from 'quill';
import Delta from 'quill-delta';
import TableCell from './js/TableCellBlot';
import TableRow from './js/TableRowBlot';
import TableHistory from './js/TableHistory';
import Table from './js/TableBlot';
import Contain from './js/ContainBlot';
import TableTrick from './js/TableTrick';
import TableSelection from './js/TableSelection';
import TableToolbar from './js/TableToolbar';
import './css/quill.table.css';

let Container = Quill.import('blots/container');

const Parchment = Quill.import('parchment');

const nodeListToArray = collection => {
  const elementsIndex = [];
  for (let i = 0; i < collection.length; i++) {
    elementsIndex.push(i);
  }
  return elementsIndex.map(i => collection.item(i));
};

Container.order = [
  'list', 'contain',   // Must be lower
  'td', 'tr', 'table'  // Must be higher
];

export default class TableModule {
  static register() {
    Quill.register(TableCell);
    Quill.register(TableRow);
    Quill.register(Table);
    Quill.register(Contain);
  }

  constructor(quill, options) {
    quill.history.tableStack = {};
    quill.table = {
      isInTable: false,
      tables: {}
    };

    // selection mouse events
    quill.container.addEventListener('mousedown', (e) => TableSelection.mouseDown(quill, e));
    quill.container.addEventListener('mousemove', (e) => TableSelection.mouseMove(quill, e));
    quill.container.addEventListener('mouseup', (e) => TableSelection.mouseUp(quill, e));
    quill.on('selection-change', (range, oldRange) => TableSelection.selectionChange(quill, range, oldRange));

    const toolbar = quill.getModule('toolbar');
    toolbar.addHandler('table', function (value) {
      return TableTrick.table_handler(value, quill);
    });

    const clipboard = quill.getModule('clipboard');
    clipboard.addMatcher('TABLE', function (node, delta) {
      const is_pasted_data = node.closest('.ql-editor') === null;
      const table_id = node.getAttribute('table_id');
      if (table_id) {
        quill.table.tables[table_id] = {
          pasted: is_pasted_data,
          row_counter: node.querySelectorAll('tr').length,
          cell_counter: node.querySelectorAll('td').length
        };
      }
      return delta;
    });
    clipboard.addMatcher('TR', function (node, delta) {
      return delta;
    });
    clipboard.addMatcher('TD', function (node, delta) {
      const tableNode = node.closest('table');
      if (!node.getAttribute('table_id') && tableNode) {
        if (!tableNode.getAttribute('table_id')) {
          tableNode.setAttribute('table_id', TableTrick.random_id());
        }
        node.setAttribute('table_id', tableNode.getAttribute('table_id'));
      }

      if (!node.getAttribute('row_id')) {
        const rowNode = node.closest('tr');
        if (rowNode) {
          if (!rowNode.getAttribute('row_id')) {
            rowNode.setAttribute('row_id', TableTrick.random_id());
          }
          node.setAttribute('row_id', rowNode.getAttribute('row_id'));
        }
      }

      if (!node.getAttribute('cell_id')) {
        node.setAttribute('cell_id', TableTrick.random_id());
      }

      return delta.compose(new Delta().retain(delta.length(), {
        td: [
          node.getAttribute('table_id'),
          node.getAttribute('row_id'),
          node.getAttribute('cell_id'),
          node.getAttribute('merge_id'),
          node.getAttribute('colspan'),
          node.getAttribute('rowspan')
        ].join('|')
      }));
    });

    TableToolbar.enable(quill, ['newtable_*', 'insert', 'undo', 'redo']);
  }

  static tableOptions() {
    const maxRows = 5;
    const maxCols = 5;
    const tableOptions = [];
    for (let r = 1; r <= maxRows; r++) {
      for (let c = 1; c <= maxCols; c++) {
        tableOptions.push('newtable_' + r + '_' + c);
      }
    }
    return tableOptions;
  }

  static removeNodeChildren(node) {
    while (node.firstChild) {
      node.removeChild(node.firstChild);
    }
  }

  static keyboardHandler(quill, key, range, keycontext) {
    const format_start = quill.getFormat(range.index - 1);
    const format_end = quill.getFormat(range.index + range.length);

    if (key === 'undo' || key === 'redo' || key === 'copy') {
      return TableTrick.table_handler(key, quill);
    }

    // If the event is not in a cell, then pass the standard handler
    if (!format_start.td && !keycontext.format.td && !format_end.td) {
      return true;
    }

    if (key === 'backspace') {
      // if the selection is at the cell border
      if (!keycontext.offset && !range.length) {
        const selection = window.getSelection();
        const nodeList = document.querySelectorAll(".ql-editor p");
        // remove selected content
        const resultNodes = nodeListToArray(nodeList).filter(cell =>
          selection.containsNode(cell, true)
        );

        // deletion does not affect the cell
        if (!resultNodes.length) {
          return true;
        }
        // do not delete if we have a selection (TODO: manage selection deletion)
        if (TableSelection.getSelectionCoords()) {
          return false;
        }

        let nodeRemoved = false;
        resultNodes.forEach((resultNode, i) => {
          if (resultNode.previousSibling) {
            if (resultNode.previousSibling.nodeName === 'TABLE') {
              // remove last cell if we are right after a table
              const cells = resultNode.previousSibling.querySelectorAll('td');
              if (cells.length && TableTrick._removeCell(cells[cells.length - 1])) {
                nodeRemoved = true;
              }
            }
          } else if (resultNode.parentNode.nodeName === 'TD') {
            // remove current cell if we are inside it
            if (TableTrick._removeCell(resultNode.parentNode)) {
              nodeRemoved = true;
            }
          }
        });

        if (nodeRemoved) {
          TableHistory.add(quill);
        }

        // if at least one node has been removed, then return false (do not call standard handler)
        return !nodeRemoved;
      }

      // If we delete not at the cell border, then pass the standard handler
      return true;
    }

    let node = quill.selection.getNativeRange().start.node;
    if (!node) return false
    let blot = Parchment.find(node);

    if (
      key === 'delete' && blot &&
      keycontext.offset < (blot.text ? blot.text.length : 0)
    ) {
      return true;
    }

    const [prev] = quill.getLine(range.index - 1);
    const [next] = quill.getLine(range.index + 1);
    // If a cell has multiple rows, you can delete as standard
    if (key === 'backspace' && prev && prev.next) {
      return true;
    }
    if (key === 'delete' && next && next.prev) {
      return true;
    }
  }
}
